from   numba import njit
import numpy as     np

class gls_periodogram:

    def gls_periodogram(self, time:list, val:list, err:list=None, ofac:int=1, normalize:bool=True, win_func:bool=False) -> tuple:
        """Generalized Lomb-Scargle (GLS) periodogram.

        :param time: time values
        :type time: list
        :param val: values
        :type val: list
        :param err: errors, defaults to None
        :type err: list, optional
        :param ofac: over-factorization, defaults to 1
        :type ofac: int, optional
        :param normalize: normalization of periodogram, defaults to True
        :type normalize: bool, optional
        :param win_func: return window function, defaults to False
        :type win_func: bool, optional
        :return: frequency, power spectrum and phase of periodogram; if win_func is True, the frequency, power spectrum and phase of the window function are returned as well
        :rtype: tuple
        """

        # if not provided, set uncertainties to unity
        if err is None:
            err = np.ones(len(time))

        # frequencies, power spectrum and phases of data
        freq, ps, phi = _gls(time, val, err, ofac, normalize)

        # spectral window function
        if win_func:

            # central frequency
            freq_c = freq[int(len(freq)/2)]

            # sinusoid with unit amplitude at central frequency
            win_val = np.sin(2*np.pi*freq_c*time)
            win_err = err

            # power spectrum of window function
            win_freq, win_ps, _ = _gls(time, win_val, win_err, ofac)

            # recenter window function frequencies
            win_freq -= freq_c

            # area of window function
            win_dfreq = np.mean(win_freq[1:]-win_freq[:-1])
            win_area = np.sum(win_ps)*win_dfreq

        # return periodogram parameters
        if win_func:
            return freq, ps, phi, win_freq, win_ps, win_area
        else:
            return freq, ps, phi

def _gls(time, val, err, ofac, normalize=True):

    # time span and steps
    Time = time[-1] - time[0]
    dtime = time[1:] - time[:-1]

    # linear and angular frequencies
    dfreq = 1/(Time*ofac)
    freq = np.arange(1/Time, 1/(2*np.median(dtime)), dfreq)
    omega = 2*np.pi*freq

    # weights
    W = np.sum(1/err**2)
    w = 1/(W*err**2)

    # empty arrays for powers and phases
    N = len(freq)
    ps = np.empty(N)
    phi = np.empty(N)

    # loop through frequencies
    for i in range(N):

        # trigonometric terms
        arg = omega[i] * time
        cosarg = np.cos(arg)
        sinarg = np.sin(arg)

        # weighted sums
        Y = _weight_sum1(w, val)
        C = _weight_sum1(w, cosarg)
        S = _weight_sum1(w, sinarg)

        # weighted sums of cross-terms
        YYhat = _weight_sum2(w, val, val)
        YChat = _weight_sum2(w, val, cosarg)
        YShat = _weight_sum2(w, val, sinarg)
        CChat = _weight_sum2(w, cosarg, cosarg)
        SShat = _weight_sum2(w, sinarg, sinarg)
        CShat = _weight_sum2(w, cosarg, sinarg)

        # differences of sums
        YY = YYhat - Y * Y
        YC = YChat - Y * C
        YS = YShat - Y * S
        CC = CChat - C * C
        SS = SShat - S * S
        CS = CShat - C * S

        # normalization
        D = CC * SS - CS ** 2

        # amplitudes
        a = (YC * SS - YS * CS) / D
        b = (YS * CC - YC * CS) / D

        # power spectrum
        if normalize == True:
            ps[i] = (SS*YC**2 + CC*YS**2 - 2*CS*YC*YS) / (YY*D)
        if normalize == False:
            ps[i] = a**2 + b**2

        # phases
        phi[i] = np.arctan2(a, b)

    # return frequencies, power spectrum and phases
    return freq, ps, phi

@njit()
def _weight_sum1(w, arr):
    sumw = 0
    for i in range(len(w)):
        sumw += w[i] * arr[i]
    return sumw

@njit()
def _weight_sum2(w, arr1, arr2):
    sumw = 0
    for i in range(len(w)):
        sumw += w[i] * arr1[i] * arr2[i]
    return sumw


class sptype_to_num:

    def sptype_to_num(self, sptype:str) -> int:
        """Spectral type to number.

        :param sptype: spectral type
        :type sptype: str
        :return: spectral type represented as a number
        :rtype: int
        """

        return "OBAFGKM".index(sptype[0])*10 + int(sptype[1])


from .add_data          import add_data
from .compute_spec_mast import compute_spec_mast
from .compute_vrad_ccf  import compute_vrad_ccf
from .compute_vrad_lbl  import compute_vrad_lbl
from .get_aux_data      import get_aux_data

from typing import Optional

class Data(
    add_data,
    compute_spec_mast,
    compute_vrad_ccf,
    compute_vrad_lbl,
    get_aux_data
    ):
    """ARVE Data sub-class.
    """

    def __init__(self, arve):
        self.arve                     = arve
        self.time    : Optional[dict] = None
        self.vrad    : Optional[dict] = None
        self.spec    : Optional[dict] = None
        self.aux_data: Optional[dict] = None


import numpy             as     np
import pandas            as     pd
from   scipy.interpolate import interp1d
import warnings
warnings.filterwarnings("ignore")

class compute_spec_mast:

    def compute_spec_mast(self, ofac=10) -> None:
        """Compute master spectrum.

        :param ofac: oversampling factor, defaults to 10
        :type ofac: int, optional
        :return: None
        :rtype: None
        """

        # read data from input
        if self.spec["path"] is None:

            # read data
            wave_val, flux_val, flux_err = [self.spec[key] for key in ["wave_val", "flux_val", "flux_err"]]

            # master spectrum
            mast_flux_val = np.average(flux_val, weights=1/flux_err**2, axis=0)
            
            # remove NaN
            mast_flux_val[np.isnan(mast_flux_val)] = 0

            # oversample
            func_flux     = interp1d(wave_val, mast_flux_val, "cubic")
            mast_wave_val = np.concatenate([np.linspace(wave_val[i], wave_val[i+1], ofac+1)[:ofac] for i in range(len(wave_val)-1)])
            mast_wave_val = np.append(mast_wave_val, wave_val[-1])
            mast_flux_val = func_flux(mast_wave_val)
        
        # read data from path
        else:

            # read data
            wave_val = self.spec["wave_val"]

            # nr. of spectra
            Nspec = len(self.spec["files"])

            # loop spectra
            for i in range(Nspec):

                # read CSV file
                df = pd.read_csv(self.spec["files"][i])
                
                # get flux and flux error if same wavelength grid
                if self.spec["same_wave_grid"]:
                    flux_val   = df["flux_val"].to_numpy()
                    flux_err   = df["flux_err"].to_numpy()
                
                # interpolate flux and flux error on reference wavelength grid
                else:
                    wave_val_i = df["wave_val"].to_numpy()
                    flux_val_i = df["flux_val"].to_numpy()
                    flux_err_i = df["flux_err"].to_numpy()
                    flux_val   = interp1d(wave_val_i, flux_val_i, kind="cubic", bounds_error=False)(wave_val)
                    flux_err   = interp1d(wave_val_i, flux_err_i, kind="cubic", bounds_error=False)(wave_val)
                
                # master spectrum
                if i == 0:
                    mast_flux_val = flux_val
                    mast_flux_err = flux_err
                else:
                    flux_val_arr = np.array([mast_flux_val,flux_val])
                    flux_err_arr = np.array([mast_flux_err,flux_err])
                    mast_flux_val = np.average(flux_val_arr, weights=1/flux_err_arr**2, axis=0)
                    mast_flux_err = np.sqrt(1/np.sum(1/flux_err_arr**2, axis=0))

            # remove NaN
            mast_flux_val[np.isnan(mast_flux_val)] = 0
            
            # oversample
            func_flux     = interp1d(wave_val, mast_flux_val, "cubic")
            mast_wave_val = np.concatenate([np.linspace(wave_val[i], wave_val[i+1], ofac+1)[:ofac] for i in range(len(wave_val)-1)])
            mast_wave_val = np.append(mast_wave_val, wave_val[-1])
            mast_flux_val = func_flux(mast_wave_val)

        # save spectral data
        self.spec_mast = {"wave_val": mast_wave_val, "flux_val": mast_flux_val}

        return None


class doppler_shift:

    def doppler_shift(self, wave:list, v:float) -> list:
        """Doppler shift.

        :param wave: wavelengths
        :type wave: list
        :param v: velocity in km/s
        :type v: float
        :return: Doppler-shifted wavelengths
        :rtype: list
        """

        # read constants
        c = self.constants["c"] # [km/s] speed of light in vacuum

        return wave*((1+v/c)/(1-v/c))**(1/2)


import numpy             as     np
import pandas            as     pd
from   scipy.interpolate import interp1d
from   scipy.optimize    import curve_fit
from   tqdm              import tqdm
import warnings
warnings.filterwarnings('ignore')

class compute_vrad_lbl:

    def compute_vrad_lbl(self, mask_path:str=None, weight:str=None, criteria:list=None, exclude_tellurics=True, scale=True, Niter=1) -> None:
        """Compute radial velocities (RVs) from spectral data.

        :param mask_path: path to line mask (must be a CSV file where the wavelength column is "wave"), defaults to None
        :type mask_path: str, optional
        :param weight: column name of weight, defaults to None
        :type weight: str, optional
        :param criteria: criteria to apply (must be columns with prefix "crit_"), defaults to None
        :type criteria: list, optional
        :param exclude_tellurics: exclude telluric bands, defaults to True
        :type exclude_tellurics: bool, optional
        :param scale: scale master spectrum when fitting, defaults to True
        :type scale: bool, optional
        :param Niter: number of iterations per line, defaults to 1
        :type Niter: int, optional
        :return: None
        :rtype: None
        """

        # read data
        wave_val                     =  self.spec["wave_val"]
        mast_wave_val, mast_flux_val = [self.spec_mast[key] for key in ["wave_val", "flux_val"]]
        if self.spec["path"] is None:
            flux_val_arr, flux_err_arr = [self.spec[key] for key in ["flux_val", "flux_err"]]

        # read constants
        c = self.arve.functions.constants["c"]

        # mask from VALD
        if mask_path is None:

            # read mask and mask name
            mask      = self.aux_data["mask"]
            mask_name = self.aux_data["name"]
        
        # provided mask
        else:

            # read mask and mask name
            mask      = pd.read_csv(mask_path)
            mask_name = mask_path.split("/")[-1]

        # central, left and right wavelengths
        wc = np.array(mask["wave"  ])
        wl = np.array(mask["wave_l"])
        wr = np.array(mask["wave_r"])

        # keep mask lines which satisfy criteria
        if criteria is not None:
            idx = np.ones_like(wc, dtype=bool)
            for i in range(len(criteria)):
                crit = np.array(mask["crit_"+criteria[i]])
                idx *= crit
            wc  = wc[idx]
            wl  = wl[idx]
            wr  = wr[idx]

        # exclude tellurics
        if exclude_tellurics:

            # read telluric bands
            tell      = self.aux_data["tell"]
            wave_l    = np.array(tell["wave_l"])
            wave_r    = np.array(tell["wave_r"])
            wave_band = np.array([wave_l,wave_r]).T

            # keep lines outside of bands
            idx = np.sum([(wc > wave_band[i,0]) & (wc < wave_band[i,1]) for i in range(len(wave_band))], axis=0).astype(bool) == False
            wc  = wc[idx]
            wl  = wl[idx]
            wr  = wr[idx]
        
        # nr. of spectra
        if self.spec["path"] is None:
            Nspec = len(flux_val_arr)
        else:
            Nspec = len(self.spec["files"])

        # nr. of lines
        Nline = len(wc)

        # empty arrays for RV values and errors
        vrad_val_arr = np.zeros((Nspec,Nline))
        vrad_err_arr = np.zeros((Nspec,Nline))

        # master spectrum gradient
        mast_grad_val = np.gradient(mast_flux_val)/np.gradient(mast_wave_val)
        
        # flux and gradient functions
        flux_func  = interp1d(mast_wave_val, mast_flux_val, kind="cubic")
        grad_func  = interp1d(mast_wave_val, mast_grad_val, kind="cubic")

        # empty array for line indices
        ilr = np.zeros(Nline, dtype=object)

        # loop lines
        for i in range(Nline):
        
            # line index range in wavelength grid
            il     = np.searchsorted(wave_val, wl[i], side="right")
            ir     = np.searchsorted(wave_val, wr[i], side="right")
            ilr[i] = np.arange(il, ir)

        # loop spectra
        print("Analyzed spectra:")
        for i in tqdm(range(Nspec)):

            # read data from input
            if self.spec["path"] is None:
                flux_val = flux_val_arr[i]
                flux_err = flux_err_arr[i]
            
            # read data from path
            else:

                # read CSV file
                df = pd.read_csv(self.spec["files"][i])
                
                # get flux and flux error if same wavelength grid
                if self.spec["same_wave_grid"]:
                    flux_val   = df["flux_val"].to_numpy()
                    flux_err   = df["flux_err"].to_numpy()
                
                # interpolate flux and flux error on reference wavelength grid
                else:
                    wave_val_i = df["wave_val"].to_numpy()
                    flux_val_i = df["flux_val"].to_numpy()
                    flux_err_i = df["flux_err"].to_numpy()
                    flux_val   = interp1d(wave_val_i, flux_val_i, kind="cubic", bounds_error=False)(wave_val)
                    flux_err   = interp1d(wave_val_i, flux_err_i, kind="cubic", bounds_error=False)(wave_val)

            # loop lines
            for j in range(Nline):

                # initial guess
                vrad_val = 0

                # wavelength, flux and flux uncertainty for line
                wave_spec_line = wave_val[ilr[j]]
                flux_spec_line = flux_val[ilr[j]]
                ferr_spec_line = flux_err[ilr[j]]

                # try to template match
                try:

                    # loop iterations
                    for _ in range(Niter):

                        # single spectrum shift
                        wave_spec_shift = wave_spec_line*(1-vrad_val/c)
                        flux_spec_shift = flux_spec_line
                        ferr_spec_shift = ferr_spec_line
                        
                        # master spectrum interpolate
                        wave_mast_inter = wave_spec_shift
                        flux_mast_inter = flux_func(wave_mast_inter)
                        grad_mast_inter = grad_func(wave_mast_inter)

                        # select function of shifted spectrum and initial guess on parameters
                        if scale:
                            func_spec_shift = _spec_shift_scale
                            p0              = (0,0,1)
                        else:
                            func_spec_shift = _spec_shift
                            p0              = (0,)

                        # solve for parameters
                        param, _ = curve_fit(func_spec_shift, (wave_mast_inter, flux_mast_inter, grad_mast_inter), flux_spec_shift, sigma=ferr_spec_shift, p0=p0)

                        # compute RV and RV error
                        vrad_val += param[0]*c
                        vrad_err  = 1/np.sqrt(np.sum(1/(ferr_spec_shift/(grad_mast_inter*wave_mast_inter/c))**2))

                    # save
                    vrad_val_arr[i,j] = vrad_val
                    vrad_err_arr[i,j] = vrad_err
                
                # if unsuccessful, assign NaN to RV and RV error
                except:

                    vrad_val_arr[i,j] = np.nan
                    vrad_err_arr[i,j] = np.nan

        # weighted average of valid lines
        vrad_val = np.zeros(Nspec)
        vrad_err = np.zeros(Nspec)
        for i in range(Nspec):
            idx = ~np.isnan(vrad_val_arr[i,:]) & (np.abs(vrad_err_arr[i,:])>0)
            vrad_val[i] = np.average(vrad_val_arr[i,idx], weights=1/vrad_err_arr[i,idx]**2)
            vrad_err[i] = np.sqrt(1/np.sum(1/vrad_err_arr[i,idx]**2))

        # save RV data
        self.vrad = {"vrad_val": vrad_val, "vrad_err": vrad_err, "vrad_val_lbl": vrad_val_arr, "vrad_err_lbl": vrad_err_arr, "method": "LBL", "mask": mask_name}

        return None

# spectrum shifted
def _spec_shift(X, *param):

    wave, flux, grad = X
    z = param
    S = flux - grad*wave*z

    return S

# spectrum shfited and scaled
def _spec_shift_scale(X, *param):

    wave, flux, grad = X
    z, c0, c1 = param
    S = c0 + c1*(flux - grad*wave*z)

    return S


from .convert_air_to_vac import convert_air_to_vac
from .convert_vac_to_air import convert_vac_to_air
from .doppler_shift      import doppler_shift
from .gls_periodogram    import gls_periodogram
from .inverted_gaussian  import inverted_gaussian
from .sptype_to_num      import sptype_to_num

from typing import Optional

class Functions(
    convert_air_to_vac,
    convert_vac_to_air,
    doppler_shift,
    gls_periodogram,
    inverted_gaussian,
    sptype_to_num
    ):
    """ARVE Functions sub-class.
    """

    def __init__(self, arve):
        self.arve            = arve
        self.constants: dict = {
            "c": 2.99792458e5 # [km/s] speed of light in vacuum
        }


from .add_vpsd_components    import add_vpsd_components
from .compute_vpsd           import compute_vpsd
from .fit_vpsd_coefficients  import fit_vpsd_coefficients
from .get_stellar_parameters import get_stellar_parameters
from .plot_vpsd_components   import plot_vpsd_components

from typing import Optional

class Star(
    add_vpsd_components,
    compute_vpsd,
    fit_vpsd_coefficients,
    get_stellar_parameters,
    plot_vpsd_components
    ):
    """ARVE Star sub-class.
    """

    def __init__(self, arve):
        self.arve                               = arve
        self.target            : Optional[str]  = None
        self.stellar_parameters: Optional[dict] = None
        self.vpsd              : Optional[dict] = None
        self.vpsd_components   : Optional[dict] = None


import numpy             as     np
import pandas            as     pd
from   scipy.interpolate import interp1d
from   scipy.optimize    import curve_fit
from   tqdm              import tqdm

class compute_vrad_ccf:

    def compute_vrad_ccf(self, mask_path:str=None, weight:str=None, criteria:list=None, exclude_tellurics=True, vgrid:list=[-20,20,0.25]) -> None:
        """Compute radial velocities (RVs) from spectral data.

        :param mask_path: path to line mask (must be a CSV file where the wavelength column is "wave"), defaults to None
        :type mask_path: str, optional
        :param weight: column name of weight, defaults to None
        :type weight: str, optional
        :param criteria: criteria to apply (must be columns with prefix "crit_"), defaults to None
        :type criteria: list, optional
        :param exclude_tellurics: exclude telluric bands, defaults to True
        :type exclude_tellurics: bool, optional
        :param vgrid: velocity grid, in the format [start,stop,step] and in units of km/s, on which to evaluate the CCF, defaults to [-20,20,0.25]
        :type vgrid: list, optional
        :return: None
        :rtype: None
        """

        # read data
        wave_val = self.spec["wave_val"]
        if self.spec["path"] is None:
            flux_val_arr, flux_err_arr = [self.spec[key] for key in ["flux_val", "flux_err"]]

        # mask from VALD
        if mask_path is None:

            # read mask and mask name
            mask      = self.aux_data["mask"]
            mask_name = self.aux_data["name"]
        
        # provided mask
        else:

            # read mask and mask name
            mask      = pd.read_csv(mask_path)
            mask_name = mask_path.split("/")[-1]

        # central wavelengths
        wc = np.array(mask["wave"])

        # weights
        if weight is None:
            w = np.ones_like(wc)
        else:
            w = np.array(mask[weight])

        # keep mask lines which satisfy criteria
        if criteria is not None:
            idx = np.ones_like(wc, dtype=bool)
            for i in range(len(criteria)):
                crit = np.array(mask["crit_"+criteria[i]])
                idx *= crit
            wc  = wc[idx]
            w   = w [idx]

        # exclude tellurics
        if exclude_tellurics:

            # read telluric bands
            tell      = self.aux_data["tell"]
            wave_l    = np.array(tell["wave_l"])
            wave_r    = np.array(tell["wave_r"])
            wave_band = np.array([wave_l,wave_r]).T

            # keep lines outside of bands
            idx = np.sum([(wc > wave_band[i,0]) & (wc < wave_band[i,1]) for i in range(len(wave_band))], axis=0).astype(bool) == False
            wc  = wc[idx]
            w   = w [idx]

        # RV shifts
        vrads = np.arange(vgrid[0],vgrid[1]+vgrid[2]/2,vgrid[2])

        # keep mask lines within spectrum overlap
        idx = (self.arve.functions.doppler_shift(wave=wc, v=min(vrads)) > min(wave_val)) & \
              (self.arve.functions.doppler_shift(wave=wc, v=max(vrads)) < max(wave_val))
        wc  = wc[idx]
        w   = w [idx]

        # normalize weights
        w = w/np.sum(w)
        
        # nr. of spectra
        if self.spec["path"] is None:
            Nspec = len(flux_val_arr)
        else:
            Nspec = len(self.spec["files"])
        
        # nr. of RV shifts
        Nvrad = len(vrads)
        
        # nr. of lines
        Nline = len(wc)

        # empty arrays for indices
        ir = np.zeros((Nvrad,Nline), dtype=int)
        il = np.zeros((Nvrad,Nline), dtype=int)

        # empty arrays for fractions
        fr = np.zeros((Nvrad,Nline))
        fl = np.zeros((Nvrad,Nline))

        # loop RV shifts
        for i in range(Nvrad):

            # shifted central wavelengths
            wc_shift = self.arve.functions.doppler_shift(wave=wc, v=vrads[i])

            # right and left indices
            ir[i] = np.searchsorted(wave_val, wc_shift, side="right")
            il[i] = ir[i] - 1

            # right and left fractions
            fr[i] = (wc_shift-wave_val[il[i]])/(wave_val[ir[i]]-wave_val[il[i]])
            fl[i] = 1 - fr[i]

        # empty arrays for RV values and errors
        vrad_val = np.zeros(Nspec)
        vrad_err = np.zeros(Nspec)

        # empty arrays for FWHM values and errors
        fwhm_val = np.zeros(Nspec)

        # loop spectra
        print("Analyzed spectra:")
        for i in tqdm(range(Nspec)):

            # read data from input
            if self.spec["path"] is None:
                flux_val = flux_val_arr[i]
                flux_err = flux_err_arr[i]
            
            # read data from path
            else:

                # read CSV file
                df = pd.read_csv(self.spec["files"][i])
                
                # get flux and flux error if same wavelength grid
                if self.spec["same_wave_grid"]:
                    flux_val   = df["flux_val"].to_numpy()
                    flux_err   = df["flux_err"].to_numpy()
                
                # interpolate flux and flux error on reference wavelength grid
                else:
                    wave_val_i = df["wave_val"].to_numpy()
                    flux_val_i = df["flux_val"].to_numpy()
                    flux_err_i = df["flux_err"].to_numpy()
                    flux_val   = interp1d(wave_val_i, flux_val_i, kind="cubic", bounds_error=False)(wave_val)
                    flux_err   = interp1d(wave_val_i, flux_err_i, kind="cubic", bounds_error=False)(wave_val)

            # empty arrays for CCF values and errors
            ccf_val = np.zeros(Nvrad)
            ccf_err = np.zeros(Nvrad)

            # loop RV shifts
            for j in range(Nvrad):

                # CCF value and error
                ccf_val[j] = np.nansum((flux_val[il[j]]   *fl[j]+flux_val[ir[j]]   *fr[j])*w   )
                ccf_err[j] = np.nansum((flux_err[il[j]]**2*fl[j]+flux_err[ir[j]]**2*fr[j])*w**2)**(1/2)

            # initial guess on Gaussian parameters
            i_min = np.argmin(ccf_val)
            i_max = np.argmax(ccf_val)
            C0 = ccf_val[i_max]
            a0 = 1-ccf_val[i_min]/C0
            b0 = vrads[i_min]
            c0 = (b0-interp1d(ccf_val[:i_min], vrads[:i_min], kind="cubic")((ccf_val[i_min]+ccf_val[i_max])/2))*2
            p0 = (C0, a0, b0, c0)

            # CCF parameters with fitted Gaussian
            param, _ = curve_fit(self.arve.functions.inverted_gaussian, vrads, ccf_val, sigma=ccf_err, p0=p0)

            # RV value and error
            vrad_val[i] = param[2]
            vrad_err[i] = 1/np.sqrt(np.sum(1/np.abs(ccf_err*np.gradient(vrads)/np.gradient(ccf_val))**2))

            # FWHM value and error
            fwhm_val[i] = param[3]

        # save RV data
        self.vrad = {"vrad_val": vrad_val, "vrad_err": vrad_err, "method": "CCF", "mask": mask_name}
        self.ccf  = {"fwhm_val": fwhm_val}

        return None


from .data      import Data
from .functions import Functions
from .planets   import Planets
from .star      import Star

import gc
import pickle

from typing import Optional

class ARVE:
    """ARVE main class.
    """

    def __init__(self):
        self.id       : Optional[str] = None
        self.data     : Data          = Data(self)
        self.functions: Functions     = Functions(self)
        self.planets  : Planets       = Planets(self)
        self.star     : Star          = Star(self)

def save(arve:ARVE) -> None:
    """Save ARVE object.

    :param arve: ARVE object to save
    :type arve: ARVE
    :return: None
    :rtype: None
    """

    return pickle.dump(arve, open(arve.id+'.arve', 'wb'))

def load(arve:str) -> ARVE:
    """Load ARVE object.

    :param arve: ARVE file to load
    :type arve: str
    :return: loaded ARVE object
    :rtype: ARVE
    """

    return pickle.load(open(arve, 'rb'))

def delete(arve:ARVE) -> None:
    """Delete ARVE object.

    :param arve: ARVE object to delete
    :type arve: ARVE
    :return: None
    :rtype: None
    """
    
    del arve
    gc.collect()
    
    return None


import numpy as np

class compute_vpsd:

    def compute_vpsd(self) -> None:
        """Compute velocity power spectral density (VPSD).

        :return: None
        :rtype: None
        """

        # read RV time series
        time, vrad_val, vrad_err, time_unit, vrad_unit = [self.arve.data.vrad[key] for key in ["time", "vrad_val", "vrad_err", "time_unit", "vrad_unit"]]

        # compute velocity power spectrum
        freq, vps, phi, win_freq, win_vps, win_area = self.arve.functions.gls_periodogram(time=time, val=vrad_val, err=vrad_err, normalize=False, win_func=True)

        # compute log-average VPS
        freq_bin = 10 ** (np.linspace(np.log10(freq[0]), np.log10(freq[-1]), 51))
        freq_avg = (freq_bin[1:] + freq_bin[:-1]) / 2
        vps_avg  = np.empty(freq_avg.size)
        for i in range(freq_avg.size):
            vps_bin = vps[(freq > freq_bin[i]) & (freq < freq_bin[i + 1])]
            if len(vps_bin) == 0:
                vps_avg[i] = np.nan
            else:
                vps_avg[i] = np.mean(vps_bin)

        # delete NaN values
        i_delete = np.isnan(vps_avg)
        freq_avg = np.delete(freq_avg, np.where(i_delete))
        vps_avg  = np.delete(vps_avg , np.where(i_delete))

        # compute VPSD
        vpsd     = vps    /win_area
        vpsd_avg = vps_avg/win_area

        # save VPSD
        self.vpsd = {"freq": freq, "vps": vps, "vpsd": vpsd, "phase": phi, "freq_avg": freq_avg, "vps_avg": vps_avg, "vpsd_avg": vpsd_avg}

        return None


class convert_air_to_vac:

    def convert_air_to_vac(self, wave_air:list) -> list:
        """Convert wavelengths from air to vacuum.

        :param wave_air: wavelengths in air [Å]
        :type wave_air: list
        :return: wavelengths in vacuum [Å]
        :rtype: list
        """

        s = 1e4 / wave_air
        n = 1 + 0.00008336624212083 + 0.02408926869968 / (130.1065924522 - s**2) + 0.0001599740894897 / (38.92568793293 - s**2)
        
        wave_vac = wave_air * n

        return wave_vac


import numpy                 as     np
import os
import pandas                as     pd
import pkg_resources
from   scipy.interpolate     import interp1d
from   scipy.ndimage.filters import convolve

class get_aux_data:

    def get_aux_data(self) -> None:
        """Get auxiliary data.

        :return: None
        :rtype: None
        """
        
        # read wavelengh range
        wave_val = self.spec["wave_val"]
        wave_min = min(wave_val)
        wave_max = max(wave_val)

        # read resolution and medium
        resolution = self.spec["resolution"]
        medium     = self.spec["medium"]

        # path to auxiliary data
        path_aux_data = pkg_resources.resource_filename("arve", "aux_data/")

        # search masks
        masks            = os.listdir(path_aux_data+"masks/")
        masks            = [mask for mask in masks if mask.endswith(".csv.zip")]
        sptype_masks     = [mask.split(".")[0] for mask in masks]
        sptype_num_masks = np.array([self.arve.functions.sptype_to_num(sptype=sptype) for sptype in sptype_masks])

        # search spec
        specs            = os.listdir(path_aux_data+"spectra/")
        specs            = [spec for spec in specs if spec.endswith(".csv.zip")]
        sptype_specs     = [spec.split(".")[0] for spec in specs]
        sptype_specs.remove('WAVE')
        sptype_num_specs = np.array([self.arve.functions.sptype_to_num(sptype=sptype) for sptype in sptype_specs])

        # spectral type as number
        sptype_num       = self.arve.functions.sptype_to_num(sptype=self.arve.star.stellar_parameters["sptype"])

        # read closest mask
        idx_mask = np.argmin(np.abs(sptype_num-sptype_num_masks))
        mask     = pd.read_csv(path_aux_data+"masks/"+masks[idx_mask])
        if medium == "air":
            mask["wave"  ] = self.arve.functions.convert_vac_to_air(mask["wave"  ])
            mask["wave_l"] = self.arve.functions.convert_vac_to_air(mask["wave_l"])
            mask["wave_r"] = self.arve.functions.convert_vac_to_air(mask["wave_r"])
        idx_wave = (mask.wave_l>=wave_min) & (mask.wave_r<=wave_max)
        mask     = mask[idx_wave].reset_index(drop=True)

        # read closest spec
        idx_spec = np.argmin(np.abs(sptype_num-sptype_num_specs))
        spec     = pd.read_csv(path_aux_data+"spectra/"+specs[idx_spec])
        wave     = pd.read_csv(path_aux_data+"spectra/WAVE.csv.zip")["wave"]
        if medium == "air":
            wave = self.arve.functions.convert_vac_to_air(wave)
        spec.insert(0, "wave", wave)

        # interpolate spectrum
        spec_new         = pd.DataFrame()
        spec_new["wave"] = wave_val
        spec_new["flux"] = interp1d(spec.wave, spec.flux, kind="cubic")(spec_new.wave)
        spec_new["temp"] = interp1d(spec.wave, spec.temp, kind="cubic")(spec_new.wave)
        spec             = spec_new

        # convolve spectrum
        if resolution is not None:
            spec["flux"] = _convolve_gaussian(spec.wave.to_numpy(), spec.flux.to_numpy(), resolution)

        # read tellurics
        tell     = pd.read_csv(path_aux_data+"tellurics/TELL.csv.zip")
        tell["wave_l"] = self.arve.functions.doppler_shift(tell["wave_l"], -self.arve.star.stellar_parameters["vrad_sys"])
        tell["wave_r"] = self.arve.functions.doppler_shift(tell["wave_r"], -self.arve.star.stellar_parameters["vrad_sys"])
        if medium == "air":
            tell["wave_l"] = self.arve.functions.convert_vac_to_air(tell["wave_l"])
            tell["wave_r"] = self.arve.functions.convert_vac_to_air(tell["wave_r"])
        idx_wave = (tell.wave_l>=wave_min) & (tell.wave_r<=wave_max)
        tell     = tell[idx_wave].reset_index(drop=True)

        # find left and right index of lines
        mask.insert(mask.columns.get_loc("wave_l")+2, "wave_l_idx", np.searchsorted(spec.wave, mask.wave_l))
        mask.insert(mask.columns.get_loc("wave_r")+2, "wave_r_idx", np.searchsorted(spec.wave, mask.wave_r))
        mask["wave_l"    ] = spec.wave[mask.wave_l_idx].to_numpy()
        mask["wave_r"    ] = spec.wave[mask.wave_r_idx].to_numpy()

        # save
        self.aux_data = {"name": masks[idx_mask],
                         "mask": mask,
                         "spec": spec,
                         "tell": tell}

        return None

def _convolve_gaussian(wave:list, flux:list, resolution:float) -> list:
    """Convolve spectrum with Gaussian instrumental profile.

    :param wave: wavelength values
    :type wave: list
    :param flux: flux values
    :type flux: list
    :param resolution: instrumental resolution
    :type resolution: float
    :return: fluxed convolved with Gaussian instrumental profile
    :rtype: list
    """

    # nr. of wavelength points
    Nwave = len(wave)

    # half-width at half-maximum
    hwhm = 0.5*wave[0]/resolution

    # uniform dispersion
    wave_ran = wave[-1]-wave[0]
    wave_del = wave_ran/(Nwave-1)
    
    # nr. of points in half Gaussian
    Nhalf = int(4/np.sqrt(2*np.log(2))*hwhm/wave_del)
    
    # nr. of points in fill Gaussian (odd)
    Nfull = 2*Nhalf+1
    
    # wavelength grid of Gaussian
    wave_grid = wave_del*(np.arange(Nfull)-(Nfull-1)/2)
    
    # standard deviation
    sigma = hwhm/np.sqrt(2*np.log(2))
    
    # normalized Gaussian
    gauss  = np.exp(-(wave_grid/sigma)**2/2)
    gauss /= np.sum(gauss)

    # convolve spectrum
    flux = convolve(flux, gauss, mode="nearest")

    return flux


import numpy as np

class add_vpsd_components:

    def add_vpsd_components(self, components:list=["Photon_noise", "Oscillations", "Granulation", "Supergranulation"]) -> None:
        """Add velocity power spectral density (VPSD) components.

        :param components: VPSD components, defaults to ["Photon_noise", "Oscillations", "Granulation", "Supergranulation"]
        :type components: list, optional
        :return: None
        :rtype: None
        """

        # read stellar parameters and VPSD
        Teff, logg, M, R = [self.stellar_parameters[key] for key in ["Teff", "logg", "M", "R"]]
        freq, vpsd, freq_avg, vpsd_avg = [self.vpsd[key] for key in ["freq", "vpsd", "freq_avg", "vpsd_avg"]]

        # photon noise
        if "Photon_noise" in components:

            # coefficients
            c0 = np.min(vpsd_avg) # amplitude

            # specifications
            name     = "Photon_noise" # name
            type     = "Constant"     # function type
            coef_val = [c0  ]         # coefficient values
            coef_err = [0   ]         # coefficient errors
            vary     = [True]         # vary when fitted

            # check noise is non-negative
            if c0 >= 0:

                # save VPSD component
                self.vpsd_components[name] = {"type": type, "coef_val": coef_val, "coef_err": coef_err, "vary": vary}

        # oscillations
        if "Oscillations" in components:

            # coefficients
            c2 = M/(R**2*np.sqrt(Teff/5777))*3.05e-3 * 60*60*24 # central frequency
            c1 = M**(1/2)*R**(-3/2)*134.9*1e-6*2 * 60*60*24     # peak width
            c0 = vpsd_avg[np.argmin(np.abs(freq_avg-c2))]       # amplitude

            # specifications
            name     = "Oscillations"     # name
            type     = "Lorentz"          # function type
            coef_val = [c0  , c1  , c2  ] # coefficient values
            coef_err = [0   , 0   , 0   ] # coefficient errors
            vary     = [True, True, True] # vary when fitted

            # check central frequency is within frequency range
            if (c2 > min(freq)) & (c2 < max(freq)):

                # save VPSD component
                self.vpsd_components[name] = {"type": type, "coef_val": coef_val, "coef_err": coef_err, "vary": vary}

        # granulation
        if "Granulation" in components:

            # coefficients
            c2 = 2                                                    # exponent
            c1 = 1/24*(10**logg/10**(4.4))**(7/9)*(Teff/5777)**(23/9) # characteristic timescale
            c0 = vpsd_avg[np.argmin(np.abs(freq_avg-1/c1))]           # amplitude

            # specifications
            name     = "Granulation"       # name
            type     = "Harvey"            # function type
            coef_val = [c0  , c1  , c2   ] # coefficient values
            coef_err = [0   , 0   , 0    ] # coefficient errors
            vary     = [True, True, False] # vary when fitted

            # check characteristic frequency is within frequency range
            if (1/c1 > min(freq)) & (1/c1 < max(freq)):

                # save VPSD component
                self.vpsd_components[name] = {"type": type, "coef_val": coef_val, "coef_err": coef_err, "vary": vary}

        # supergranulation
        if "Supergranulation" in components:

            # coefficients
            c2 = 2                                                     # exponent
            c1 = 10/24*(10**logg/10**(4.4))**(7/9)*(Teff/5777)**(23/9) # characteristic timescale
            c0 = vpsd_avg[np.argmin(np.abs(freq_avg-1/c1))]            # amplitude

            # specifications
            name     = "Supergranulation"  # name
            type     = "Harvey"            # function type
            coef_val = [c0  , c1  , c2   ] # coefficient values
            coef_err = [0   , 0   , 0    ] # coefficient errors
            vary     = [True, True, False] # vary when fitted

            # check characteristic frequency is within frequency range
            if (1/c1 > min(freq)) & (1/c1 < max(freq)):

                # save VPSD component
                self.vpsd_components[name] = {"type": type, "coef_val": coef_val, "coef_err": coef_err, "vary": vary}
        
        return None


from .add_planet import add_planet

from typing import Optional

class Planets(
    add_planet
    ):
    """ARVE Planets sub-class.
    """
    
    def __init__(self, arve):
        self.arve                       = arve
        self.parameters: Optional[dict] = None


import matplotlib.pyplot as plt
import numpy             as np

class plot_vpsd_components:

    def plot_vpsd_components(self) -> plt.Figure:
        """Plot velocity power spectral density (VPSD) components.

        :return: figure with the VPSD, its log-average and its modeled components
        :rtype: plt.Figure
        """

        # figure
        fig = plt.figure()

        # read VPSD and units
        freq, vpsd, freq_avg, vpsd_avg = [self.vpsd[key] for key in ["freq", "vpsd", "freq_avg", "vpsd_avg"]]
        time_unit, vrad_unit = [self.arve.data.vrad[key] for key in ["time_unit", "vrad_unit"]]

        # plot VPSD and average VPSD
        plt.loglog(freq, vpsd, ls="-", c="k", alpha=0.5)
        plt.loglog(freq_avg, vpsd_avg, ls="None", marker="o", mec="k", mfc="None")

        # empty array for sum of components
        vpsd_tot = np.zeros(len(freq))

        # loop components
        for comp in self.vpsd_components.keys():

            # component dictionary
            comp_dict = self.vpsd_components[comp]

            # type and coefficients
            type     = comp_dict["type"]
            coef_val = comp_dict["coef_val"]

            # type Constant
            if type == "Constant":

                # unpack coefficients
                c0 = coef_val[0]

                # compute component
                vpsd_comp = c0

                # plot component
                plt.axhline(vpsd_comp, ls="--", label=comp)

            # type Lorentz
            if type == "Lorentz":
                
                # unpack coefficients
                c0 = coef_val[0]
                c1 = coef_val[1]
                c2 = coef_val[2]

                # compute component
                vpsd_comp = c0*c1**2/(c1**2+(freq-c2)**2)

                # plot component
                plt.loglog(freq, vpsd_comp, ls="--", label=comp)

            # type Harvey
            if type == "Harvey":

                # unpack coefficients
                c0 = coef_val[0]
                c1 = coef_val[1]
                c2 = coef_val[2]

                # compute component
                vpsd_comp = c0/(1+(c1*freq)**c2)

                # plot component
                plt.loglog(freq, vpsd_comp, ls="--", label=comp)

            # add component to sum
            vpsd_tot += vpsd_comp

        # plot component sum
        plt.loglog(freq, vpsd_tot, ls="-", c="k", label="Total")

        # plot limits
        plt.xlim(freq[0], freq[-1])

        # plot labels
        plt.xlabel("$f$" + " " + f"[{time_unit}" + "$^{-1}$]")
        plt.ylabel("VPSD" + " " + f"[({vrad_unit})" + "$^{2}$" + " / " + f"{time_unit}" + "$^{-1}$]")

        # plot legend
        leg = plt.legend(loc="lower left")
        leg.set_zorder(101)

        # plot layout
        plt.tight_layout()

        # return figure
        return fig


class convert_vac_to_air:

    def convert_vac_to_air(self, wave_vac:list) -> list:
        """Convert wavelengths from vacuum to air.

        :param wave_vac: wavelengths in vacuum [Å]
        :type wave_vac: list
        :return: wavelengths in air [Å]
        :rtype: list
        """

        s = 1e4 / wave_vac
        n = 1 + 0.0000834254 + 0.02406147 / (130 - s**2) + 0.00015998 / (38.9 - s**2)
        
        wave_air = wave_vac / n

        return wave_air


import numpy as     np
from   lmfit import Parameters, minimize

class fit_vpsd_coefficients:

    def fit_vpsd_coefficients(self) -> None:
        """Fit velocity power spectral density (VPSD) coefficients.

        :return: None
        :rtype: None
        """

        # read VPSD
        freq, vpsd, freq_avg, vpsd_avg = [self.vpsd[key] for key in ["freq", "vpsd", "freq_avg", "vpsd_avg"]]

        # LMFIT parameters
        params = Parameters()

        # loop components
        for comp in self.vpsd_components.keys():

            # component dictionary
            comp_dict = self.vpsd_components[comp]

            # coefficients and vary
            coef_val = comp_dict["coef_val"]
            vary     = comp_dict["vary"]

            # loop coefficients
            for i in range(len(coef_val)):

                # add parameters
                params.add(comp + "_" + str(i), value=coef_val[i], min=coef_val[i]/10, max=coef_val[i]*10, vary=vary[i])

        # fit coefficients
        c = minimize(_func_res, params, args=(self, freq_avg, vpsd_avg))

        # loop components
        for comp in self.vpsd_components.keys():

            # component dictionary
            comp_dict = self.vpsd_components[comp]

            # coefficients
            coef_val = comp_dict["coef_val"]
            coef_err = comp_dict["coef_err"]

            # loop coefficients
            for i in range(len(coef_val)):

                # update coefficients with fitted values
                coef_val[i] = c.params[comp + "_" + str(i)].value
                coef_err[i] = c.params[comp + "_" + str(i)].stderr
        
        return None

def _func_res(params, self, freq_avg, vpsd_avg):

    # empty array for sum of components
    vpsd_tot = np.zeros(len(freq_avg))

    # loop components
    for comp in self.vpsd_components.keys():

        # component dictionary
        comp_dict = self.vpsd_components[comp]

        # component type
        type = comp_dict["type"]

        # type Constant
        if type == "Constant":
            
            # unpack coefficients
            c0 = params[comp + "_0"]

            # compute component
            vpsd_comp = c0

        # type Lorentz
        if type == "Lorentz":

            # unpack coefficients
            c0 = params[comp + "_0"]
            c1 = params[comp + "_1"]
            c2 = params[comp + "_2"]

            # compute component
            vpsd_comp = c0*c1**2/(c1**2+(freq_avg-c2)**2)

        # type Harvey
        if type == "Harvey":

            # unpack coefficients
            c0 = params[comp + "_0"]
            c1 = params[comp + "_1"]
            c2 = params[comp + "_2"]

            # compute component
            vpsd_comp = c0/(1+(c1*freq_avg)**c2)
    
        # add component to sum
        vpsd_tot += vpsd_comp
    
    # logarithmic residual
    logres = np.log10(vpsd_avg) - np.log10(vpsd_tot)

    # return residual
    return logres


from   astroquery.simbad import Simbad
import numpy             as     np

class get_stellar_parameters:

    def get_stellar_parameters(self) -> None:
        """Get spectral type from SIMBAD query and stellar parameters from main-sequence table.

        :return: None
        :rtype: None
        """

        # initiate dictionarty with stellar parameters
        self.stellar_parameters = {}

        # Sun
        if self.target == "Sun":

            # save stellar parameters
            self.stellar_parameters["sptype"  ] = "G2"
            self.stellar_parameters["vrad_sys"] = 0.0
            self.stellar_parameters["Teff"    ] = 5770
            self.stellar_parameters["logg"    ] = 4.4
            self.stellar_parameters["Fe_H"    ] = 0.0
            self.stellar_parameters["M"       ] = 1.0
            self.stellar_parameters["R"       ] = 1.0
            self.stellar_parameters["vsini"   ] = 1.63

        # other stars
        else:

            # get spectral type from query
            simbad = Simbad()
            simbad.add_votable_fields("sptype"  )
            simbad.add_votable_fields("velocity")
            self.stellar_parameters["sptype"  ] = simbad.query_object(self.target)["SP_TYPE"   ][0][:2]
            self.stellar_parameters["vrad_sys"] = simbad.query_object(self.target)["RVZ_RADVEL"][0]

            # convert spectral types to numbers
            sptype_num       =  self.arve.functions.sptype_to_num(sptype=self.stellar_parameters["sptype"])
            sptype_num_table = [self.arve.functions.sptype_to_num(sptype=sptype) for sptype in _table["sptype"]]

            # interpolate and save stellar parameters from table
            for key in ["Teff", "logg", "Fe_H", "M", "R", "vsini"]:
                self.stellar_parameters[key] = np.interp(sptype_num, sptype_num_table, _table[key])
            
        # compute and save micro- and macro-turbulence
        self.stellar_parameters["vmic"] = 0.85
        self.stellar_parameters["vmac"] = max(0.00, 3.98-(self.stellar_parameters["Teff"]-5770)/650)

        return None

# table with spectral parameters for main sequence stars
_table = np.array(
[
("A0", 9572, 4.3, 0.0, 2.34, 1.80, 255.0),
("A2", 8985, 4.3, 0.0, 2.21, 1.75, 244.0),
("A5", 8306, 4.2, 0.0, 2.04, 1.69, 225.0),
("A7", 7935, 4.2, 0.0, 1.93, 1.68, 210.0),
("F0", 7178, 4.3, 0.0, 1.66, 1.62, 180.0),
("F2", 6909, 4.3, 0.0, 1.56, 1.48, 135.0),
("F5", 6528, 4.3, 0.0, 1.41, 1.40,  20.0),
("F8", 6160, 4.4, 0.0, 1.25, 1.20,   9.0),
("G0", 5943, 4.4, 0.0, 1.16, 1.12,   6.4),
("G2", 5811, 4.4, 0.0, 1.11, 1.08,   4.8),
("G5", 5657, 4.5, 0.0, 1.05, 0.95,   3.4),
("G8", 5486, 4.5, 0.0, 0.97, 0.91,   2.6),
("K0", 5282, 4.6, 0.0, 0.90, 0.83,   2.2),
("K2", 5055, 4.6, 0.0, 0.81, 0.75,   2.0),
("K3", 4973, 4.6, 0.0, 0.79, 0.73,   2.0),
("K5", 4623, 4.6, 0.0, 0.65, 0.64,   1.9),
("K7", 4380, 4.7, 0.0, 0.54, 0.54,   1.7),
("M0", 4212, 4.7, 0.0, 0.46, 0.48,   1.5),
("M2", 4076, 4.7, 0.0, 0.40, 0.43,   0.0),
("M5", 3923, 4.8, 0.0, 0.34, 0.38,   0.0)
],
dtype=[("sptype","U2"), ("Teff","f4"), ("logg","f4"), ("Fe_H","f4"), ("M","f4"), ("R","f4"), ("vsini","f4")]
)


import glob
import pandas as pd

class add_data:

    def add_data(
        self,
        time_val      :list=None,
        vrad_val      :list=None,
        vrad_err      :list=None,
        wave_val      :list=None,
        flux_val      :list=None,
        flux_err      :list=None,
        resolution    :int =None,
        medium        :str="vac",
        path          :str= None,
        compression   :str= None,
        same_wave_grid:bool=False
        ) -> None:
        """Add data.

        :param time_val: time values, defaults to None
        :type time_val: list, optional
        :param vrad_val: radial velocity values, defaults to None
        :type vrad_val: list, optional
        :param vrad_err: radial velocity errors, defaults to None
        :type vrad_err: list, optional
        :param wave_val: wavelength values, defaults to None
        :type wave_val: list, optional
        :param flux_val: flux values, defaults to None
        :type flux_val: list, optional
        :param flux_err: flux errors, defaults to None
        :type flux_err: list, optional
        :param resolution: instrumental resolution, defaults to None
        :type resolution: int, optional
        :param medium: medium, defaults to "vac"
        :type medium: str, optional
        :param path: path to spectra which must be stored in CSV files, defaults to None
        :type path: str, optional
        :param compression: extension ('zip', 'gzip', etc) if the CSV files are compressed, defaults to None
        :type compression: str, optional
        :param same_wave_grid: all spectra already on the same wavelength grid, defaults to False
        :type same_wave_grid: bool, optional
        :return: None
        :rtype: None
        """
        
        # input from path
        if path is not None:
            
            # path to CSV files
            path_csv = path+'**/*.csv'

            # compressed extension
            if compression is not None:
                path_csv += '.'+compression
            
            # search files
            files = glob.glob(path_csv, recursive=True)
            files = sorted(files)

            # add first spectrum as reference
            df       = pd.read_csv(files[0])
            wave_val = df["wave_val"].to_numpy()
            flux_val = df["flux_val"].to_numpy()
            flux_val = flux_val.reshape(1,len(flux_val))
            flux_err = df["flux_err"].to_numpy()
            flux_err = flux_err.reshape(1,len(flux_err))
        
        else:

            # None entry for files
            files = None

        # add dictionaries with data
        self.time = {"time_val"      : time_val}
        self.vrad = {"vrad_val"      : vrad_val,
                     "vrad_err"      : vrad_err}
        self.spec = {"wave_val"      : wave_val,
                     "flux_val"      : flux_val,
                     "flux_err"      : flux_err,
                     "resolution"    : resolution,
                     "medium"        : medium,
                     "path"          : path,
                     "same_wave_grid": same_wave_grid,
                     "files"         : files}
        
        return None


from setuptools import find_packages, setup

with open("README.md") as file:
    long_description = file.read()

setup(
    name="arve",
    version="0.1.7",
    author="Khaled Al Moulla",
    author_email="khaled.almoulla@gmail.com",
    description="Analyzing Radial Velocity Elements",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/almoulla/arve",
    license="MIT License",
    packages=find_packages(),
    install_requires=["astroquery",
                      "lmfit"     ,
                      "matplotlib",
                      "numba"     ,
                      "numpy"     ,
                      "pandas"    ,
                      "scipy"     ,
                      "tqdm"      ,
                     ],
    classifiers=["Development Status :: 1 - Planning"    ,
                 "Intended Audience :: Science/Research" ,
                 "License :: OSI Approved :: MIT License",
                 "Programming Language :: Python :: 3"   ,
                 ],
    include_package_data=True,
    package_data={"arve": ["aux_data/masks/*.csv.zip",
                           "aux_data/spectra/*.csv.zip",
                           "aux_data/tellurics/*.csv.zip"]}
)


import numpy as np

class inverted_gaussian:

    def inverted_gaussian(self, x:list, *params:tuple) -> list:
        """Inverted Gaussian.

        :param x: RV array
        :type x: list
        :param params: tuple with continuum, contrast, RV and FWHM
        :type params: tuple of floats
        :return: inverted Gaussian evaluated at x
        :rtype: list
        """
        
        # unpack parameters
        continuum, contrast, vrad, fwhm = params
        
        # rename parameters
        C = continuum
        a = contrast
        b = vrad
        c = fwhm
        
        # scale FWHM into sigma
        c /= 2*np.sqrt(2*np.log(2))

        return C*(1-a*np.exp(-((x-b)/c)**2/2))